#include <iostream>
#include "main.h"

#ifndef _WIN32
#define _WIN32
#endif

#define UCI 1
#define FILE_CNT

#define UseDebug 0

//  
U64 hash_key [13][64];  // 
U64 hash_key_ep [64];   //   
U64 hash_key_cast [16]; //   
U64 hash_key_wtm;       //  

int   material    [12] = {b_p, w_p, b_n, w_n, b_b, w_b, b_r, w_r, b_q, w_q, b_k, w_k};
char  piece_array [13] = "pPnNbBrRqQkK";


// King in middle game
int pst_km     [] = {-3,-4,-4,-5,-5,-4,-4,-3,
                     -3,-4,-4,-5,-5,-4,-4,-3,
                     -3,-4,-4,-5,-5,-4,-4,-3,
                     -3,-4,-4,-5,-5,-4,-4,-3,
                     -2,-3,-3,-4,-4,-3,-3,-2,
                     -1,-2,-2,-2,-2,-2,-2,-1,
                      2, 2, 0, 0, 0, 0, 2, 2,
                      2, 3, 1, 0, 0, 1, 3, 2};

// King in end game
int pst_ke     [] = {-5,-4,-3,-2,-2,-3,-4,-5,
                     -3,-2,-1, 0, 0,-1,-2,-3,
                     -3,-1, 2, 3, 3, 2,-1,-3,
                     -3,-1, 3, 4, 4, 3,-1,-3,
                     -3,-1, 3, 4, 4, 3,-1,-3,
                     -3,-1, 2, 3, 3, 2,-1,-3,
                     -3,-3, 0, 0, 0, 0,-3,-3,
                     -5,-3,-3,-3,-3,-3,-3,-5};

// Piece Square Table (PST)
int pst [12][64];

//  
int n_trop [64][64];
int b_trop [64][64];
int r_trop [64][64];
int q_trop [64][64];

//   
int un_castling [64] = {0,};

int history [64][64];
int killers [80][2];

int king_cast[2] = {c_k, c_K};
int queen_cast[2] = {c_q, c_Q};

pv_t pv [PV_SIZE];

int nodes [80];


hash * H = new hash;
board * _B = new board [MAX_DEPTH];
search * S = new search;
reps_t * R = new reps_t [REP_TABLE];

board * last = NULL;

int     g_hash_age;

int     wtm;
U64     piece [12];
U64     occupied [2];
int     squares [64];   //    , ,   
int     mat [2];       //     

int     err = 0;
FILE*   f_out    = NULL;
FILE*   f_magics = NULL;

magic_t bishop_m [];
magic_t rook_m  [];

U64     bishop_att [][512] = {{0,},};
U64     rook_att  [][4096] = {{0,},};

U64     knight_att [] = {0,};
U64     king_att [] = {0,};
U64     link_att [][64] = {{0,},};

U64     wpawn_att [] = {0,};
U64     bpawn_att [] = {0,};

int main ()
{
    input_available ();

    board * B = &_B[0];

    srand (1502314);
    g_hash_age = 0;

    //setvbuf (stdin,NULL,_IONBF,0); // Causes ERRORS! Don't touch! This is not a joke!
    //setvbuf (stdout,NULL,_IONBF,0); // _IOLBF breaks on Windows!

    f_out    = fopen (F_REPORT, "wb");
	f_magics = fopen (F_MAGICS, "rb");

    init_attacks ();
    init_magics ();
	init_pst ();

    eval_init ();
    hash_init ();
    board_init ();
    search_preinit ();
	search_init ();

	//      (GUI  )
	uci_send ("%s v%s by %s\n", NAME, VERSION, AUTHOR);

	/*
	print_this_att (c_att_K);
	print_this_att (c_att_Q);
	print_this_att (c_att_k);
	print_this_att (c_att_q);

	print_this_att (c_mov_K);
	print_this_att (c_mov_Q);
	print_this_att (c_mov_k);
	print_this_att (c_mov_q);
	*/

	//for (i = 0; i < 64; i ++) print_this_att (queen_att [i]);


	//uci_send ("%lld\n", 111099948471);



	/*//  
	for (i = 0; i < 90; i ++)
	{
		occ = rand64();

		att = r_att (occ, D4); //     D4

		print_this_att (occ);
		print_this_att (att);

		att = b_att (occ, D4); //     D4

		print_this_att (att);

		uci_send ("\n_______________\n\n");
	};
	//*/


	//   "uci"
	uci_wait ();

	srand ((unsigned)time(NULL));

	//  
	fen_to_pos (START_FEN, B);

	hash_calc (B);

	//     GUI
	for (;;) if (uci_parsing (B)) break;

	//getchar ();
	//getchar ();

    return 0;
};

void board_init ()
{
    board * B;

    for (int i = 0; i < MAX_DEPTH; i ++)
    {
        B = &_B[i];

        B->check_is = 0;
        B->attacker = 0;
        B->castling_rights = 15;
        B->en_passant = 0;
        B->hash = E;
        B->hash_mat = E;
        B->taken = no_p;

        B->next = &_B[i + 1];

        last = B;
    };

    // ,   
    _B[MAX_DEPTH - 1].next = &_B[0];


    for (int j = 0; j < 12; j ++) piece [j] = E;
    occupied [0] = occupied [1] = E;
    for (int j = 0; j < 64; j ++) squares [j] = no_p;
};

void search_preinit ()
{
    S->irreversible_ply = 0;
    S->fifty = 0;
	S->count_moves = 0;
};

void search_init ()
{
    int i;

    S->state = ANALYSING;
    S->infinite = 0;
    S->bm = S->tm = 0;
    S->score = 0;
    S->score_last = 0;
    S->best_move = 0;

    for (int i = 0; i < 80; i ++) nodes [i] = 0;
    S->in_pv = 0;

    S->hash_probe = 0;
    S->hash_count = 0;

    S->curr_time = 0;

    S->think_time = S->elaps_time = 0;

    S->pv_updated = 0;
    S->pv_len = 0;
    S->pv_start = 0;

    S->nodes_count = 0;

    S->ply = 0;
    S->search_depth = 0;

    for (i = 0; i < PV_SIZE; i ++)
	{
	    pv [i].key = E;
	    pv [i].key_new = E;
        pv [i].mv = 0;
	};
};


void hash_init ()
{
    int i, j;

	for (i = 0; i < 12; i ++)
	{
		for (j = 0; j < 64; j ++)
		{
			hash_key [i][j] = rand64();
		};
	};

	for (i = A3; i <= H3; i ++) hash_key_ep [i] = rand64();
	for (i = A6; i <= H6; i ++) hash_key_ep [i] = rand64();

	for (i = 0; i < 16; i ++) hash_key_cast [i] = rand64();

	hash_key_wtm = rand64();

	hash_refresh ();
	reps_refresh ();
};

void eval_init ()
{
    int i, j;
    U64 bb;

    for (i = 0; i < 64; i ++) un_castling [i] = c_k | c_q | c_K | c_Q;

	un_castling [A1] = c_k | c_q | c_K       ;
	un_castling [H1] = c_k | c_q       | c_Q ;
	un_castling [A8] = c_k       | c_K | c_Q ;
	un_castling [H8] =       c_q | c_K | c_Q ;

	un_castling [E1] = c_k | c_q             ;
	un_castling [E8] =             c_K | c_Q ;

	// 

	for (i = 0; i < 64; i ++) //  
	{
	    tropism_calc (i, knight_id);
	    tropism_calc (i, bishop_id);
	    tropism_calc (i, rook_id);
	    tropism_calc (i, queen_id);
	};

    /*
    //   
    int x, y;
    int xk, yk;

	for (i = 0; i < 64; i ++)
	{
	    xk = vert (i);
	    yk = hori (i);

	    for (y = 7; y >= 0; y --)
	    {
	        for (x = 0; x < 8; x ++)
	        {
	            j = compose (y, x);
                if (i == j) printf (" -K-");
                else 	    printf ("%3i ", q_trop [i][j]); //   
	        };

	        printf ("\n");
	    };

	    printf ("\n");
	};
	*/
};

void tropism_calc (int king, int piece)
{
    int j, val, ply;
    U64 att, bb, t;
    U64 saved_att;
    U64 dont_looked;

    const int ntrop [] = {0, 10, 8, 5, 2, 1, 0, -1, -2, -3};
    const int btrop [] = {0, 10, 4, 0, 0};
    const int rtrop [] = {0, 8, 7, 0, 0};
    const int qtrop [] = {0, 7, 2, 0, 0};

    att = I << king; //    0
    dont_looked = U ^ att;

    for (ply = 1; ply < 20; ply ++)
    {
        bb = att; //   ply - 1
        att = E; //   ply

        //   
        while (bb)
        {
            t = LSB (bb);
            j = bitscan (t);

            switch (piece)
            {
                case knight_id : att |= knight_att [j]; break;
                case bishop_id : att |= b_att (E, j);   break;
                case rook_id   : att |= r_att (E, j);   break;
                case queen_id  : att |= q_att (E, j);   break;
            };

            bb ^= t;
        };

        att &= dont_looked; //   ...
        dont_looked ^= att; //                   ... 

        saved_att = att;

        while (att)
        {
            t = LSB (att);
            j = bitscan (t);

            //    
            switch (piece)
            {
                case knight_id : n_trop [king][j] = ntrop [ply];                         break;
                case bishop_id : b_trop [king][j] = btrop [ply] - (dist (king, j) >> 2); break;
                case rook_id   : r_trop [king][j] = rtrop [ply] - (dist (king, j) >> 1); break;
                case queen_id  : q_trop [king][j] = qtrop [ply] - (dist (king, j) >> 2); break;
            };

            att ^= t;
        };

        att = saved_att;
    };

    //  
    switch (piece)
    {
        case knight_id : n_trop [king][king] = 0; break;
        case bishop_id : b_trop [king][king] = 0; break;
        case rook_id   : r_trop [king][king] = 0; break;
        case queen_id  : q_trop [king][king] = 0; break;
    };
};

bool input_available()
{
    static bool init = false, is_pipe;
    static HANDLE stdin_h;
    DWORD val, error;

    val = 0; // needed to make the compiler happy?

    // have a look at the "local" buffer first, *this time before init (no idea if it helps)*

    if (UseDebug && !init) printf("info string init=%d stdin->_cnt=%d\n",int(init),stdin->_cnt);

    if (stdin->_cnt > 0) return true; // HACK: assumes FILE internals

    // input init (only done once)

    if (!init)
    {
        init = true;

        stdin_h = GetStdHandle(STD_INPUT_HANDLE);

        if (UseDebug && (stdin_h == NULL || stdin_h == INVALID_HANDLE_VALUE))
        {
            error = GetLastError();
            printf("info string GetStdHandle() failed, error=%d\n",error);
        }

        is_pipe = !GetConsoleMode(stdin_h,&val); // HACK: assumes pipe on failure

        if (UseDebug) printf("info string init=%d is_pipe=%d\n",int(init),int(is_pipe));

        if (UseDebug && is_pipe)
        {
            // GetConsoleMode() failed, everybody assumes pipe then
            error = GetLastError();
            printf("info string GetConsoleMode() failed, error=%d\n",error);
        }

        if (!is_pipe)
        {
            SetConsoleMode (stdin_h,val & ~(ENABLE_MOUSE_INPUT|ENABLE_WINDOW_INPUT));
            FlushConsoleInputBuffer (stdin_h); // no idea if we can lose data doing this
        }
    }

    // different polling depending on input type
    // does this code work at all for pipes?

    if (is_pipe)
    {

        if (!PeekNamedPipe(stdin_h, NULL, 0, NULL, &val, NULL))
        {

            if (UseDebug)
            {
                // PeekNamedPipe() failed, everybody assumes EOF then
                error = GetLastError();
                printf("info string PeekNamedPipe() failed, error=%d\n", error);
            }

            return true; // HACK: assumes EOF on failure
        }

        if (UseDebug && val < 0) printf("info string PeekNamedPipe(): val=%d\n", val);

        return (val > 0); // != 0???

    }
    else
    {
        GetNumberOfConsoleInputEvents (stdin_h, &val);
        return (val > 1); // no idea why 1
    }

    return false;
}
