// +----------------------------------------------------------------------------------+
// |      main.h - header file, written by Nikita Kurgin (aka neurocore), 2012        |
// +----------------------------------------------------------------------------------+

#include <stdio.h>
#include <stdlib.h>
#include <cstring>
#include <ctime>
#include <cstdarg>
#include <windows.h>
#include <conio.h>

#define C64(a) (a##ULL)

typedef          long long  S64;
typedef unsigned long long  U64;
typedef unsigned int        U32;
typedef unsigned char     uchar;

///////////////////////////////////////////
/*
#ifdef WIN64
#endif

#ifdef WIN32
#endif
*/
///////////////////////////////////////////

#define NAME      "Eia"
#define VERSION   "0.1 w32"
#define AUTHOR    "Nikita Kurgin"

#define F_REPORT  "report.txt"
#define F_MAGICS  "magics.txt"

enum   {MAX_DEPTH        = 100,
        GO_MAX_DEPTH     = 98,
        BENCH_LW_DEPTH   = 1,
        BENCH_HG_DEPTH   = 6,
        ROOT_ORDERING    = 12,              //      
        SINGULAR_VAL     = 220,            //      
        DEPTH_TO_MOVE    = 3,             //  ,    
        USE_NULL_MOVE    = 1,            //    ?
        HASH_SIZE        = 1 << 20,     //     (  )
        PV_SIZE          = 1 << 16,    //     
        REP_TABLE        = 5900,      //   
        CHECK_PLY_LIMIT  = 7,        //    
        PPUSH_PLY_LIMIT  = 6,       //    
        NODES_OF_SILENCE = 1 << 14,
        NODES_OF_OUTPUT  = 1 << 19,
        DEPTH_TO_CURRMV  = 6,    //     currmove  
        USE_LAZY         = 1,   //    ? (   )
        GEN_MAGICS       = 0}; //     ?

//     (LERF-mapping)

enum   {E        = C64(0x0000000000000000),
        I        = C64(0x0000000000000001),
        U        = C64(0xFFFFFFFFFFFFFFFF),

        file_a   = C64(0x0101010101010101),
        file_b   = C64(0x0202020202020202),
        file_c   = C64(0x0404040404040404),
        file_d   = C64(0x0808080808080808),
        file_e   = C64(0x1010101010101010),
        file_f   = C64(0x2020202020202020),
        file_g   = C64(0x4040404040404040),
        file_h   = C64(0x8080808080808080),

        rank_1   = C64(0x00000000000000FF),
        rank_2   = C64(0x000000000000FF00),
        rank_3   = C64(0x0000000000FF0000),
        rank_4   = C64(0x00000000FF000000),
        rank_5   = C64(0x000000FF00000000),
        rank_6   = C64(0x0000FF0000000000),
        rank_7   = C64(0x00FF000000000000),
        rank_8   = C64(0xFF00000000000000),

        debruijn = C64(0x07EDD5E59A4E28C2)};

const int index_bsc [64]  =  {63,  0, 58,  1, 59, 47, 53,  2,
                              60, 39, 48, 27, 54, 33, 42,  3,
                              61, 51, 37, 40, 49, 18, 28, 20,
                              55, 30, 34, 11, 43, 14, 22,  4,
                              62, 57, 46, 52, 38, 26, 32, 41,
                              50, 36, 17, 19, 29, 10, 13, 21,
                              56, 45, 25, 31, 35, 16,  9, 12,
                              44, 24, 15,  8, 23,  7,  6,  5};

//  , 

enum    {BP, WP,
		 BN, WN,
		 BB, WB,
		 BR, WR,
		 BQ, WQ,
		 BK, WK,
		 no_p = 255};

enum    {A1, B1, C1, D1, E1, F1, G1, H1,
	 	 A2, B2, C2, D2, E2, F2, G2, H2,
	 	 A3, B3, C3, D3, E3, F3, G3, H3,
	 	 A4, B4, C4, D4, E4, F4, G4, H4,
	 	 A5, B5, C5, D5, E5, F5, G5, H5,
	 	 A6, B6, C6, D6, E6, F6, G6, H6,
	 	 A7, B7, C7, D7, E7, F7, G7, H7,
	 	 A8, B8, C8, D8, E8, F8, G8, H8};


enum    {WAITING, ANALYSING, STATE_EXIT}; //    (state)
enum    {HASH_EXACT, HASH_ALPHA, HASH_BETA}; // 

#define START_FEN "rnbqkbnr/pppppppp/8/8/8/8/PPPPPPPP/RNBQKBNR w KQkq - 0 1"


//  (__|___)

enum   {f_empty    = 0,
        f_capture  = 4,
        f_castling = 8,
        f_prom     = 16,
        f_ep       = 32,
        f_2pawn    = 64,
        f_check    = 128,

        f_mask     = 252, // 111111|00

        f_color    = 1,
        BLACK      = 0,
        WHITE      = 1};

//    

enum   {o_pv        =  0x08000000,
        o_hash      =  0x04000000,
        o_prom      =  0x02000000,
        o_killer    =  0x00001000,
        o_tactical  =  0x00000010,
        o_good_cap  =  0x01000000,
        o_bad_cap   = -0x00010000,
        o_evasion   =  0x00000008,
        o_non_cap   =  0x00000000,

        o_pawn_4    =  0x00100000,
        o_pawn_5    =  0x00200000,
        o_pawn_6    =  0x00400000};

//   (   )

#define w_p   100
#define w_b   325
#define w_n   325
#define w_r   500
#define w_q   975
#define w_k   10000

#define b_p   100
#define b_b   325
#define b_n   325
#define b_r   500
#define b_q   975
#define b_k   10000

#define mate  32000
#define inf   32767

#define _k1  C64(0x5555555555555555)
#define _k2  C64(0x3333333333333333)
#define _k4  C64(0x0f0f0f0f0f0f0f0f)
#define _kf  C64(0x0101010101010101)

//   

//  ()   
#define safe_squares_b (I << G8 | I << H8 | I << A8 | I << B8 | I << C8 | I << H7 | I << G7 | I << H6)
#define safe_squares_w (I << G1 | I << H1 | I << A1 | I << B1 | I << C1 | I << H2 | I << G2 | I << H3)

//  up_mask  dn_mask
#define upmask (U ^ rank_1)
#define dnmask (U ^ rank_8)

// 
#define tempo        15

//   2 
#define bishop_pair  15

//    
#define blocked_pawn 12

//    
#define isolated     23

//     -  
#define linked_rooks 2

//    
#define lost_k_cast  10

//    
#define lost_q_cast  10

//     
#define busy_cast    3

//     
#define king_safety  3

//    
#define on_target    8

//     
#define on_king      4

//   
#define pin_bonus    30
#define ppin_bonus   12
#define disc_check   35
#define disc_att     7


//           
#define knight_on_2pawn   24
#define knight_on_pawn    16

//  
#define center_control     5
#define big_center_control 4


//     
#define shld_miss_f  30
#define shld_miss_g  30
#define shld_miss_h  22

#define shld_miss_a  22
#define shld_miss_b  30
#define shld_miss_c  30

#define shld_hole    10

//    
#define pawns_defended 3

//     
#define blocked 15

//     
#define rook_on_opened 22
#define rook_on_semi   12

//    
#define closed_king 3

//    ,   
#define uncastled 15

//    
#define pawn_on_opp_side 15

//  
#define pawn_id      0
#define knight_id    1
#define bishop_id    2
#define rook_id      3
#define queen_id     4
#define king_id      5
#define nopiece_id   7

//   
#define c_K    1
#define c_Q    2
#define c_k    4
#define c_q    8

//   
#define c_att_K ((I << E1) | (I << F1) | (I << G1))
#define c_att_Q ((I << E1) | (I << D1) | (I << C1))
#define c_att_k ((I << E8) | (I << F8) | (I << G8))
#define c_att_q ((I << E8) | (I << D8) | (I << C8))

//   
#define c_mov_K ((I << F1) | (I << G1))
#define c_mov_Q ((I << D1) | (I << C1) | (I << B1))
#define c_mov_k ((I << F8) | (I << G8))
#define c_mov_q ((I << D8) | (I << C8) | (I << B8))





//  ///////////////////////////////////////////////////////////////////////////////////////

#define hori(a)           ( (a) >> 3 )
#define vert(a)           ( (a) & 7 )
#define compose(a, b)     ( ((a) << 3) | b )
#define min(a, b)         ( (a) < (b) ? (a) : (b) )
#define max(a, b)         ( (a) > (b) ? (a) : (b) )

//  
#define absu(a, b)        ( (a) > (b) ? (a) - (b) : (b) - (a) )

//  
#define is_capture(a)     ( (a) & f_capture  )
#define is_castling(a)    ( (a) & f_castling )
#define is_prom(a)        ( (a) & f_prom     )
#define is_ep(a)          ( (a) & f_ep       )
#define is_2pawn(a)       ( (a) & f_2pawn    )
#define is_check(a)       ( (a) & f_check    )

//    (0 - , 1 - , 2 - , 3 - , 4 - , 5 - , 7 - )
#define piece_id(a)       ( ((a) & 14) >> 1  )
#define piece_color(a)    (  (a) & f_color   )
#define is_white(a)       (  ((a) & f_color) )
#define is_black(a)       ( !((a) & f_color) )

//    
#define LSB(a)            ( (U64)((U64)(a) & (U64)(E - (a))) )
#define RLSB(a)           ( (U64)((U64)(a) & (U64)((a) - 1)) )

#define bitscan(a)        ( index_bsc[( LSB(a) * debruijn ) >> 58] )

#define r_att(a, sq)      ( (U64) rook_att   [sq][( ( (U64)(a) &  rook_m[sq].mask  ) *  rook_m[sq].magic  ) >> (64 - 12)] )
#define b_att(a, sq)      ( (U64) bishop_att [sq][( ( (U64)(a) & bishop_m[sq].mask ) * bishop_m[sq].magic ) >> (64 -  9)] )
#define q_att(a, sq)      ( r_att(a, sq) | b_att(a, sq) )

// 

#define pie(a)            ( ((a) >> 24) & 0xFF )
#define from(a)           ( ((a) >> 16) & 0xFF )
#define to(a)             ( ((a) >>  8) & 0xFF )
#define flags(a)          ( (a)         & 0xFF )

//      ( !)

#define fill_up(a) ( (a) |= ((a) << 32), (a) |= ((a) << 16), (a) |= ((a) << 8) )
#define fill_dn(a) ( (a) |= ((a) >> 32), (a) |= ((a) >> 16), (a) |= ((a) >> 8) )

//  

#define generate_moves(a, b, c, d, B)   (B->moves [B->size ++] = ((a) << 24) | ((b) << 16) | ((c) << 8) | (d))
#define generate_proms(a, b, c, d, B, m) ((m) = ((a) << 24) | ((b) << 16) | ((c) << 8) | ((d) + f_prom + 0), \
                                          B->moves [B->size ++] = (m) ++, \
                                          B->moves [B->size ++] = (m) ++, \
                                          B->moves [B->size ++] = (m) ++, \
                                          B->moves [B->size ++] = (m) ++  )

// Population Count

#define pop_count(x)   (    x64 = (U64)(x),                          \
                            x64 =  x64        - ((x64 >> 1)  & _k1), \
                            x64 = (x64 & _k2) + ((x64 >> 2)  & _k2), \
                            x64 = (x64        +  (x64 >> 4)) & _k4 , \
                            (int) ((x64 * _kf) >> 56)                )

#define pop_count_few(x) pop_count(x)

//  (     )

#define r_xray(a, sq)  (  r_att (a ^ (r_att(a, sq) & a), sq) & a  )
#define b_xray(a, sq)  (  b_att (a ^ (b_att(a, sq) & a), sq) & a  )
#define q_xray(a, sq)  (  q_att (a ^ (q_att(a, sq) & a), sq) & a  )

//   

#define w_piece_mat    (  mat[1] - pop_count (piece [WP])  )
#define b_piece_mat    (  mat[0] - pop_count (piece [BP])  )

//    ()

#define dist(a, b)     (  absu(hori(a), hori(b)) + absu(vert(a), vert(b))  )

//       ?

#define mat_for_null   (piece[BR + wtm] ||                                 \
                        piece[BQ + wtm] ||                                 \
                        pop_count (piece[BN + wtm] | piece[BB + wtm]) > 1  )

//     ( -  )
#define up_mask(a)     ( upmask << ((hori(a)) << 3) )
#define dn_mask(a)     ( dnmask >> ((7 - (hori(a))) << 3) )




//  /////////////////////////////////////////////////////////////////////////////////////

struct magic_t
{
	U64 mask;
	U64 magic;
};

#pragma pack(push)
#pragma pack(1)

struct hash_t
{
	U64 key;
	int age;
	int move;
	char type;
	char depth;
	short score;
};

struct reps_t // 
{
    U64 key;
    int irreversible_last; //        
};

struct pv_t
{
    U64 key;
    U64 key_new;
    int mv;
};

#pragma pack (pop)

struct search;
struct board;

class hash;



//////////////////////////////////////////////////////////////////////////////////////////////////

class hash
{

public:

    hash_t *entry;
    int size;

    hash (int sz = HASH_SIZE)
    {
        U64 x64;

        sz /= sizeof (hash_t);
        int k = pop_count (sz); //      

        for (int i = 1; i < k; i ++) sz ^= LSB(sz);

        entry = new hash_t [sz];
        size = sz;

        //printf ("info string hash size is %i Mb now\n", (sz * sizeof (hash_t)) >> 20);

        for (int i = 0; i < sz; i ++)
        {
            entry [i].key = E;
            entry [i].age = 0;
            entry [i].move = 0;
            entry [i].type = 0;
            entry [i].depth = 0;
            entry [i].score = 0;
        };
    };

    ~hash ()
    {
        delete [] entry;
    };
};

extern hash * H; // 
extern board * _B;
extern search * S;
extern reps_t * R;


//////////////////////////////////////////////////////////////////////////////////////////////////

struct board
{
    board * next;            //   

    int check_is;          // - 
	int attacker;         //   
	int castling_rights; //    (0000QKqk)
	int en_passant;      //     

	U64 hash;
	U64 hash_mat;        //       ( )

	int taken;           //  

	U64 att_bb;          // Attacks BB
    U64 msk_bb;          // Mask BB
    U64 pin_bb;          // Pinned BB

	int size;
	int moves [239];
};


//////////////////////////////////////////////////////////////////////////////////////////////////

struct search
{
    int state;            // 
    int infinite;         //  
    int bm, tm;           // ,  
    int score;            //   
    int score_last;       //   
    int best_move;        //   

    int in_pv;

    U64 hash_probe;
    U64 hash_count;

    clock_t curr_time;

    int think_time;       //   
    int elaps_time;       //    
    int critl_time;       //  

    int is_critical;      //  ? ( ?)

    int pv_updated;       // ?
    int pv_len;           //   
    U64 pv_start;         //     pv  

    S64 nodes_count;      //   

    int ply;
    int search_depth;     //   

    int irreversible_ply; //      
    int fifty;            //    50- 
	int count_moves;      //  
};











//////     /////////////////////////////////////////////////////////

// Aging
extern int g_hash_age;

// 
extern int wtm;               //  ?
extern U64 piece [12];       //   
extern U64 occupied [2];    // 
extern int squares [64];   //   
extern int mat [2];       //     

//  
extern U64 hash_key [13][64];  // 
extern U64 hash_key_ep [64];   //   
extern U64 hash_key_cast [16]; //   
extern U64 hash_key_wtm;       //  

extern int  material       [12];
extern char piece_array    [13];

//  pst (  )
extern int pst_km[];
extern int pst_ke[];

// Piece Square Table (PST)
extern int pst [12][64];

//  
extern int n_trop [64][64];
extern int b_trop [64][64];
extern int r_trop [64][64];
extern int q_trop [64][64];

//   
extern int un_castling [64];

extern int history [64][64];
extern int killers [80][2];

extern int king_cast[2];
extern int queen_cast[2];

extern pv_t pv [PV_SIZE];

extern int nodes [80]; //     

extern magic_t bishop_m [64];
extern magic_t rook_m  [64];

extern U64 bishop_att [64][512];
extern U64 rook_att  [64][4096];

extern FILE* f_out;
extern FILE* f_magics;

extern int err;  //  

extern U64 knight_att [64];
extern U64 king_att [64];
extern U64 link_att [64][64]; //    

extern U64 wpawn_att [64];
extern U64 bpawn_att [64];








//////   ///////////////////////////////////////////////////////////////////////

extern void    hash_init              ();
extern void    hash_refresh           ();
extern hash_t* hash_get               (board * B);
extern U64     hash_calc              (board * B);
extern void    hash_update            (int mv, int score, int type, int depth, board * B);
extern void    hash_retrieve          (int &alpha, int &beta, int &mv, int depth, board * B);

///////////////////////

extern void    uci_send               (const char format[], ...);
extern void    uci_wait               ();
extern int     uci_parsing            (board * B);
extern void    uci_info               ();

///////////////////////

extern void    shell_sort             (int * a, int * b, int num);
extern int     is_on_line             (int a, int b);

//extern int     pop_count               (U64 x);
//extern int     pop_count_few           (U64 x);

/*
inline U64     fill_up                 (U64 a) {a |= (a << 32); a |= (a << 16); a |= (a << 8); return a ;}
inline U64     fill_dn                 (U64 a) {a |= (a >> 32); a |= (a >> 16); a |= (a >> 8); return a ;}
*/

inline int     random                 (int a) {return rand ()%a ;}
inline int     is_mate                (int a) {return (a < -mate || a > mate) ;}

extern void    moves_to_pos           (char * fen, board * B);

extern char*   read_command           (char * str, const char * wrd);
extern int     read_number            (char * str);
extern char*   move_to_str            (int mv);
extern int     str_to_move            (char * s);

extern U64     rand64                 ();
extern U64     rand64_fewbits         ();

//     PST
extern void    init_pst               ();
extern void    init_attacks           ();

//   
extern void    init_magics            ();
extern void    init_rook_magics       (int c);
extern void    init_bishop_magics     (int c);

extern U64     rmask                  (int sq);
extern U64     bmask                  (int sq);
extern U64     qmask                  (int sq);

extern U64     ratt                   (int sq, U64 block);
extern U64     batt                   (int sq, U64 block);

extern U64     index_to_64            (int index, int bits, U64 m);

extern void    print_pv               (int len);
extern void    sprint_pv              (int len, char* pv_line);

inline pv_t *  is_it_pv               (U64 key, int ply)
{
    int i, j = 0;
    U64 _key = S->pv_start;

    for (i = 0; i < ply; i ++)
	{
	    j = _key & (PV_SIZE - 1);
		_key = pv [j].key_new;
	};

    if (key == _key) return &pv[j];

	return NULL;
};

inline void update_kihi (int mv, int depth)
{
    int i, j;

	if (mv != killers[S->ply][0])
	{
		killers[S->ply][1] = killers[S->ply][0];
		killers[S->ply][0] = mv;
	};

	if ((history [from(mv)][to(mv)] += depth * depth) >= 60000)
	{
		for (i = 0; i < 64; i ++)
		{
			for (j = 0; j < 64; j ++)
			{
				history [i][j] >>= 1;
			};
		};
	};
};

extern U64     f_read64               (FILE* file);

extern void    print_this_att         (U64 at);
extern int     print_move             (int mv);

///////////////////////

extern void    board_init             ();

extern void    build_mask             (board * B);
extern void    check_detection        (board * B);
extern int     generate               (board * B);
extern int     generate_attacks       (board * B);
extern int     generate_legal         (board * B);
extern void    order_moves            (board * B);

extern int     make_move              (int i, board * B, board * N);
extern void    unmake_move            (int i, board * B);

extern void    make_null              (board * B, board * N);
extern void    unmake_null            ();

extern int     threefold_repetition   (board * B);

extern int     do_move                (int mv, int prom, board * B);

extern int     evaluate               (board * B);

extern void    fen_to_pos             (const char * fen, board * B);
extern char*   pos_to_fen             (board * B);

extern U64     attacked_by            (U64 occ, int sq);

extern int     print_board            (board * B);
extern void    print_moves            (board * B);

/*
inline void    generate_moves         (int a, int b, int c, int d, board * B)
{
    B->moves [B->size ++] = (a << 24) | (b << 16) | (c << 8) | d;
};

inline void    generate_proms         (int a, int b, int c, int d, board * B)
{
    int i = (a << 24) | (b << 16) | (c << 8) | (d + f_prom + 0);

    B->moves [B->size ++] = i ++;
    B->moves [B->size ++] = i ++;
    B->moves [B->size ++] = i ++;
    B->moves [B->size ++] = i ++;
};
*/



extern void    search_init            ();
extern void    search_preinit         ();
extern int     iid                    (int m_depth, board * B);
extern int     search_root            (int alpha, int beta, int depth, board * B);
extern int     search_ab              (int alpha, int beta, int depth, int in_pv, board * B);
extern int     see                    (int mv);
extern int     nega_scout             (int alpha, int beta, int depth, board * B);
extern int     q_search               (int alpha, int beta, board * B);
extern S64     perft_root             (int alpha, int beta, int depth, board * B);
extern S64     perft_ab               (int alpha, int beta, int depth, board * B);

inline void    pv_write               (board * B, board * N, int move)
{
    int ind;

    ind = B->hash & (PV_SIZE - 1);

    pv [ind].key     = B->hash;
    pv [ind].key_new = N->hash;
    pv [ind].mv      = move;
};

extern void    init_arrays            ();
extern void    init_uci_info          ();

extern bool    input_available        ();

extern void    eval_init              ();
extern void    reps_refresh           ();

inline void    print_score            (int score)
{
    uci_send ("info depth %i ", S->search_depth);

    if     (score < - mate)  uci_send ("score mate %i ", (inf + score) / 2 + 1);
    else if (score > mate)   uci_send ("score mate %i ", (inf - score) / 2 + 1);
    else                     uci_send ("score cp %i ", score);
};

extern void    tropism_calc           (int king, int piece);
